from __future__ import annotations

import re
from collections.abc import Callable
from dataclasses import dataclass, field, replace
from functools import total_ordering
from typing import Any, Self, assert_never, overload, override

from utilities.sentinel import Sentinel, sentinel
from utilities.types import MaybeCallable, MaybeStr

type VersionLike = MaybeStr[Version]
type MaybeCallableVersionLike = MaybeCallable[VersionLike]


##


@dataclass(repr=False, frozen=True, slots=True)
@total_ordering
class Version:
    """A version identifier."""

    major: int = 0
    minor: int = 0
    patch: int = 1
    suffix: str | None = field(default=None, kw_only=True)

    def __post_init__(self) -> None:
        if (self.major == 0) and (self.minor == 0) and (self.patch == 0):
            raise _VersionZeroError(
                major=self.major, minor=self.minor, patch=self.patch
            )
        if self.major < 0:
            raise _VersionNegativeMajorVersionError(major=self.major)
        if self.minor < 0:
            raise _VersionNegativeMinorVersionError(minor=self.minor)
        if self.patch < 0:
            raise _VersionNegativePatchVersionError(patch=self.patch)
        if (self.suffix is not None) and (len(self.suffix) == 0):
            raise _VersionEmptySuffixError(suffix=self.suffix)

    def __le__(self, other: Any, /) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        self_as_tuple = (
            self.major,
            self.minor,
            self.patch,
            "" if self.suffix is None else self.suffix,
        )
        other_as_tuple = (
            other.major,
            other.minor,
            other.patch,
            "" if other.suffix is None else other.suffix,
        )
        return self_as_tuple <= other_as_tuple

    @override
    def __repr__(self) -> str:
        version = f"{self.major}.{self.minor}.{self.patch}"
        if self.suffix is not None:
            version = f"{version}-{self.suffix}"
        return version

    def bump_major(self) -> Self:
        return type(self)(self.major + 1, 0, 0)

    def bump_minor(self) -> Self:
        return type(self)(self.major, self.minor + 1, 0)

    def bump_patch(self) -> Self:
        return type(self)(self.major, self.minor, self.patch + 1)

    def with_suffix(self, *, suffix: str | None = None) -> Self:
        return replace(self, suffix=suffix)


@dataclass(kw_only=True, slots=True)
class VersionError(Exception): ...


@dataclass(kw_only=True, slots=True)
class _VersionZeroError(VersionError):
    major: int
    minor: int
    patch: int

    @override
    def __str__(self) -> str:
        return f"Version must be greater than zero; got {self.major}.{self.minor}.{self.patch}"


@dataclass(kw_only=True, slots=True)
class _VersionNegativeMajorVersionError(VersionError):
    major: int

    @override
    def __str__(self) -> str:
        return f"Major version must be non-negative; got {self.major}"


@dataclass(kw_only=True, slots=True)
class _VersionNegativeMinorVersionError(VersionError):
    minor: int

    @override
    def __str__(self) -> str:
        return f"Minor version must be non-negative; got {self.minor}"


@dataclass(kw_only=True, slots=True)
class _VersionNegativePatchVersionError(VersionError):
    patch: int

    @override
    def __str__(self) -> str:
        return f"Patch version must be non-negative; got {self.patch}"


@dataclass(kw_only=True, slots=True)
class _VersionEmptySuffixError(VersionError):
    suffix: str

    @override
    def __str__(self) -> str:
        return f"Suffix must be non-empty; got {self.suffix!r}"


##


@overload
def get_version(*, version: MaybeCallableVersionLike) -> Version: ...
@overload
def get_version(*, version: None) -> None: ...
@overload
def get_version(*, version: Sentinel) -> Sentinel: ...
@overload
def get_version(
    *, version: MaybeCallableVersionLike | Sentinel
) -> Version | Sentinel: ...
@overload
def get_version(
    *, version: MaybeCallableVersionLike | None | Sentinel = sentinel
) -> Version | None | Sentinel: ...
def get_version(
    *, version: MaybeCallableVersionLike | None | Sentinel = sentinel
) -> Version | None | Sentinel:
    """Get the version."""
    match version:
        case Version() | None | Sentinel():
            return version
        case str():
            return parse_version(version)
        case Callable() as func:
            return get_version(version=func())
        case _ as never:
            assert_never(never)


##


def parse_version(version: str, /) -> Version:
    """Parse a string into a version object."""
    try:
        ((major, minor, patch, suffix),) = _PARSE_VERSION_PATTERN.findall(version)
    except ValueError:
        raise ParseVersionError(version=version) from None
    return Version(
        int(major), int(minor), int(patch), suffix=None if suffix == "" else suffix
    )


_PARSE_VERSION_PATTERN = re.compile(r"^(\d+)\.(\d+)\.(\d+)(?:-(\w+))?")


@dataclass(kw_only=True, slots=True)
class ParseVersionError(Exception):
    version: str

    @override
    def __str__(self) -> str:
        return f"Invalid version string: {self.version!r}"


__all__ = [
    "MaybeCallableVersionLike",
    "ParseVersionError",
    "Version",
    "VersionError",
    "VersionLike",
    "get_version",
    "parse_version",
]



from __future__ import annotations

from pathlib import Path
from typing import Any

from hypothesis import given
from hypothesis.strategies import booleans, floats, integers, none, text

from utilities.hypothesis import temp_paths
from utilities.pickle import read_pickle, write_pickle


class TestReadAndWritePickle:
    @given(
        obj=booleans() | integers() | floats(allow_nan=False) | text() | none(),
        root=temp_paths(),
    )
    def test_main(self, *, obj: Any, root: Path) -> None:
        write_pickle(obj, path := Path(root, "file"))
        result = read_pickle(path)
        assert result == obj



from __future__ import annotations

from numpy import allclose, isclose, linspace, pi

from utilities.statsmodels import acf


class TestACF:
    def test_main(self) -> None:
        x = linspace(0, 2 * pi, 1000)
        acfs = acf(x)
        assert acfs.shape == (31,)
        assert isclose(acfs[-1], 0.9100539400539398)

    def test_alpha(self) -> None:
        x = linspace(0, 2 * pi, 1000)
        result = acf(x, alpha=0.5)
        acfs, confint = result
        assert acfs.shape == (31,)
        assert confint.shape == (31, 2)
        assert allclose(confint[-1, :], [0.7534104755915192, 1.0666974045163604])

    def test_qstat(self) -> None:
        x = linspace(0, 2 * pi, 1000)
        result = acf(x, qstat=True)
        acfs, qstat, pvalues = result
        assert acfs.shape == (31,)
        assert qstat.shape == (30,)
        assert isclose(qstat[-1], 27769.955439016478)
        assert pvalues.shape == (30,)
        assert isclose(pvalues[-1], 0.0)

    def test_alpha_and_qstat(self) -> None:
        x = linspace(0, 2 * pi, 1000)
        result = acf(x, alpha=0.5, qstat=True)
        acfs, confint, qstat, pvalues = result
        assert acfs.shape == (31,)
        assert isclose(acfs[-1], 0.9100539400539398)
        assert confint.shape == (31, 2)
        assert allclose(confint[-1, :], [0.7534104755915192, 1.0666974045163604])
        assert qstat.shape == (30,)
        assert isclose(qstat[-1], 27769.955439016478)
        assert pvalues.shape == (30,)
        assert isclose(pvalues[-1], 0.0)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from rich.pretty import pretty_repr

from utilities.reprlib import (
    RICH_EXPAND_ALL,
    RICH_INDENT_SIZE,
    RICH_MAX_DEPTH,
    RICH_MAX_LENGTH,
    RICH_MAX_STRING,
    RICH_MAX_WIDTH,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


##


def yield_call_args_repr(
    *args: Any,
    _max_width: int = RICH_MAX_WIDTH,
    _indent_size: int = RICH_INDENT_SIZE,
    _max_length: int | None = RICH_MAX_LENGTH,
    _max_string: int | None = RICH_MAX_STRING,
    _max_depth: int | None = RICH_MAX_DEPTH,
    _expand_all: bool = RICH_EXPAND_ALL,
    **kwargs: Any,
) -> Iterator[str]:
    """Pretty print of a set of positional/keyword arguments."""
    mapping = {f"args[{i}]": v for i, v in enumerate(args)} | {
        f"kwargs[{k}]": v for k, v in kwargs.items()
    }
    return yield_mapping_repr(
        _max_width=_max_width,
        _indent_size=_indent_size,
        _max_length=_max_length,
        _max_string=_max_string,
        _max_depth=_max_depth,
        _expand_all=_expand_all,
        **mapping,
    )


##


def yield_mapping_repr(
    _max_width: int = RICH_MAX_WIDTH,
    _indent_size: int = RICH_INDENT_SIZE,
    _max_length: int | None = RICH_MAX_LENGTH,
    _max_string: int | None = RICH_MAX_STRING,
    _max_depth: int | None = RICH_MAX_DEPTH,
    _expand_all: bool = RICH_EXPAND_ALL,  # noqa: FBT001
    **kwargs: Any,
) -> Iterator[str]:
    """Pretty print of a set of keyword arguments."""
    for k, v in kwargs.items():
        v_repr = pretty_repr(
            v,
            max_width=_max_width,
            indent_size=_indent_size,
            max_length=_max_length,
            max_string=_max_string,
            max_depth=_max_depth,
            expand_all=_expand_all,
        )
        yield f"{k} = {v_repr}"


__all__ = ["yield_call_args_repr", "yield_mapping_repr"]



from __future__ import annotations

from utilities.rich import yield_call_args_repr, yield_mapping_repr


class TestYieldCallArgsRepr:
    def test_main(self) -> None:
        lines = list(yield_call_args_repr(1, 2, 3, x=4, y=5, z=6))
        expected = [
            "args[0] = 1",
            "args[1] = 2",
            "args[2] = 3",
            "kwargs[x] = 4",
            "kwargs[y] = 5",
            "kwargs[z] = 6",
        ]
        assert lines == expected


class TestYieldMappingRepr:
    def test_main(self) -> None:
        lines = list(yield_mapping_repr(a=1, b=2, c=3))
        expected = ["a = 1", "b = 2", "c = 3"]
        assert lines == expected



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, cast, overload

import statsmodels.tsa.stattools

if TYPE_CHECKING:
    from utilities.numpy import NDArrayF


type ACFMissing = Literal["none", "raise", "conservative", "drop"]


@overload
def acf(
    array: NDArrayF,
    /,
    *,
    adjusted: bool = False,
    nlags: int | None = None,
    qstat: Literal[False] = False,
    fft: bool = True,
    alpha: None = None,
    bartlett_confint: bool = True,
    missing: ACFMissing = "none",
) -> NDArrayF: ...
@overload
def acf(
    array: NDArrayF,
    /,
    *,
    adjusted: bool = False,
    nlags: int | None = None,
    qstat: Literal[False] = False,
    fft: bool = True,
    alpha: float,
    bartlett_confint: bool = True,
    missing: ACFMissing = "none",
) -> tuple[NDArrayF, NDArrayF]: ...
@overload
def acf(
    array: NDArrayF,
    /,
    *,
    adjusted: bool = False,
    nlags: int | None = None,
    qstat: Literal[True],
    fft: bool = True,
    alpha: float,
    bartlett_confint: bool = True,
    missing: ACFMissing = "none",
) -> tuple[NDArrayF, NDArrayF, NDArrayF, NDArrayF]: ...
@overload
def acf(
    array: NDArrayF,
    /,
    *,
    adjusted: bool = False,
    nlags: int | None = None,
    qstat: Literal[True],
    fft: bool = True,
    alpha: None = None,
    bartlett_confint: bool = True,
    missing: ACFMissing = "none",
) -> tuple[NDArrayF, NDArrayF, NDArrayF]: ...
@overload
def acf(
    array: NDArrayF,
    /,
    *,
    adjusted: bool = False,
    nlags: int | None = None,
    qstat: bool = False,
    fft: bool = True,
    alpha: float | None = None,
    bartlett_confint: bool = True,
    missing: ACFMissing = "none",
) -> (
    NDArrayF
    | tuple[NDArrayF, NDArrayF]
    | tuple[NDArrayF, NDArrayF, NDArrayF]
    | tuple[NDArrayF, NDArrayF, NDArrayF, NDArrayF]
): ...
def acf(
    array: NDArrayF,
    /,
    *,
    adjusted: bool = False,
    nlags: int | None = None,
    qstat: bool = False,
    fft: bool = True,
    alpha: float | None = None,
    bartlett_confint: bool = True,
    missing: ACFMissing = "none",
) -> (
    NDArrayF
    | tuple[NDArrayF, NDArrayF]
    | tuple[NDArrayF, NDArrayF, NDArrayF]
    | tuple[NDArrayF, NDArrayF, NDArrayF, NDArrayF]
):
    """Typed version of `acf`."""
    return cast(
        "Any",
        statsmodels.tsa.stattools.acf(
            array,
            adjusted=adjusted,
            nlags=nlags,
            qstat=qstat,
            fft=fft,
            alpha=alpha,
            bartlett_confint=bartlett_confint,
            missing=missing,
        ),
    )


__all__ = ["ACFMissing", "acf"]
